import re

import pytest

from eunomia.backend import BackendDict
from tests.util import temp_capture_stdout
from eunomia.config import Group, Option
from eunomia.config import keys as K
from eunomia.config import validate as V


# ========================================================================= #
# Test Config Objects                                                       #
# ========================================================================= #


def test_simple_option():
    bk = BackendDict()

    bk.dump_option(Option(data={'foo': 'bar'}))
    bk.dump_option(Option(data={'foo': 'bar'}, pkg=K.PKG_ROOT))
    bk.dump_option(Option(data={'foo': 'bar'}, pkg=K.PKG_GROUP))

    bk.dump_option(Option(data={'foo': 'bar'}, defaults=[]))

    # test relative path
    bk.dump_option(Option(data={'foo': 'bar'}, defaults=['group1/option1']))
    bk.dump_option(Option(data={'foo': 'bar'}, defaults=['group1/group2/option2']))
    # test absolute path
    bk.dump_option(Option(data={'foo': 'bar'}, defaults=['/group1/group2/option2']))

    with pytest.raises(ValueError):  # TODO: fix error messages, match='is_identifier'):
        bk.dump_option(Option(data={'foo': 'bar'}, defaults=[{'group1': '1invalid'}]))
    with pytest.raises(ValueError):  # TODO: fix error messages, match='Wrong key'):
        bk.dump_option(Option(data={'foo': 'bar'}, defaults=[{'1invalid': 'option2'}]))
    with pytest.raises(ValueError):  # TODO: fix error messages, match='Wrong key'):
        bk.dump_option(Option(data={'foo': 'bar'}, defaults=[{'group1/2invalid': 'option2'}]))

    bk.dump_option(Option(data={'foo': 'bar'}, pkg='key1'))
    bk.dump_option(Option(data={'foo': 'bar'}, pkg='key1.key2'))


def test_simple_config():
    g = Group({
        'subgroup': Group({
            'suboption2': Option({'bar': 2}),
            'suboption1': Option({'bar': 1}),
        }),
        'subgroup2': Group({
            'sub2option1': Option({'baz': 1}),
            'sub2option2': Option({'baz': 2}),
        }),
        'default': Option({
            'foo': 1
        }),
    })

    bck = BackendDict()

    # forward and backwards conversion!
    d1 = BackendDict().dump(g)
    g1 = BackendDict().load_group(d1)
    assert repr(g1) == repr(g)


def _make_config_group(suboption='suboption1', suboption2=None, package1='<group>', package2='<group>', check_cycle=False) -> Group:
    g = Group({
        'subgroup': Group({
            'suboption1': Option({'bar': 1}, pkg=package1),
            'suboption2': Option({'bar': 2}, pkg=package1),
        }),
        'subgroup2': Group({
            'subgroup3': Group({
                'sub2option1': Option({'baz': 1}, pkg=package2),
                'sub2option2': Option({'baz': 2}, pkg=package2),
            }),
        }),
        'default': Option(
            data={'foo': 1},
            defaults=[
                *([{'subgroup': suboption}] if suboption else []),  # relative
                *([{'/subgroup2/subgroup3': suboption2}] if suboption2 else []),  # absolute
            ],
        ),
    })

    if check_cycle:
        # forward and backwards conversion!
        d1 = BackendDict().dump(g)
        g1 = BackendDict().load_group(d1)
        assert repr(g1) == repr(g)

    return g


def test_config_objects():
    _make_config_group(suboption='suboption1', check_cycle=True)
    _make_config_group(suboption='suboption2', check_cycle=True)
    _make_config_group(suboption='invalid___', check_cycle=True)  # invalid but not an error


def test_option_init():
    # raw data in option
    data = {'foo': 'bar'}
    defaults = ['/group1/group2/option2']
    pkg = K.PKG_ROOT
    # make equivalent options
    option = Option(data=data, defaults=defaults, pkg=pkg)
    # check circular conversion
    bk = BackendDict()
    assert bk.dump(option) == bk.dump(bk.load_option(bk.dump(option)))

def test_debug_groups():
    root = _make_config_group(suboption='suboption1')

    with temp_capture_stdout() as out:
        root.debug_print_tree()
    color_out = out.getvalue()
    assert color_out == ' \x1b[90m\x1b[0m\x1b[35m/\x1b[0m\n \x1b[90m├\x1b[93m╌\x1b[0m \x1b[90m/:\x1b[0m \x1b[33mdefault\x1b[0m\n \x1b[90m├\x1b[95m─\x1b[0m \x1b[90m\x1b[0m\x1b[35m/subgroup\x1b[0m\n \x1b[90m│\x1b[0m  \x1b[90m├\x1b[93m╌\x1b[0m \x1b[90m/subgroup:\x1b[0m \x1b[33msuboption1\x1b[0m\n \x1b[90m│\x1b[0m  \x1b[90m╰\x1b[93m╌\x1b[0m \x1b[90m/subgroup:\x1b[0m \x1b[33msuboption2\x1b[0m\n \x1b[90m╰\x1b[95m─\x1b[0m \x1b[90m\x1b[0m\x1b[35m/subgroup2\x1b[0m\n    \x1b[90m╰\x1b[95m─\x1b[0m \x1b[90m/subgroup2\x1b[0m\x1b[35m/subgroup3\x1b[0m\n       \x1b[90m├\x1b[93m╌\x1b[0m \x1b[90m/subgroup2/subgroup3:\x1b[0m \x1b[33msub2option1\x1b[0m\n       \x1b[90m╰\x1b[93m╌\x1b[0m \x1b[90m/subgroup2/subgroup3:\x1b[0m \x1b[33msub2option2\x1b[0m\n'

    with temp_capture_stdout() as out:
        root.debug_print_tree(colors=False)
    normal_out = out.getvalue()
    assert normal_out == ' /\n ├╌ /: default\n ├─ /subgroup\n │  ├╌ /subgroup: suboption1\n │  ╰╌ /subgroup: suboption2\n ╰─ /subgroup2\n    ╰─ /subgroup2/subgroup3\n       ├╌ /subgroup2/subgroup3: sub2option1\n       ╰╌ /subgroup2/subgroup3: sub2option2\n'
    # https://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python
    assert normal_out == re.sub(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])', '', color_out)

    # TODO: test other flags

# ========================================================================= #
# END                                                                       #
# ========================================================================= #



from typing import Dict, Union, List, Tuple

from eunomia._util_traverse import RecursiveTransformer
from eunomia.config.nodes import ConfigNode, SubNode

from eunomia.config import keys as K
from eunomia.config import validate as V


# ========================================================================= #
# Config Node                                                               #
# ========================================================================= #


class _ConfigObject(object):

    def __init__(self):
        super().__init__()
        self._parent = None
        self._key = None
        self._children = {}

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Path                                                                  #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @property
    def has_parent(self):
        assert not ((not self._key) ^ (not self._parent))
        return (self._key is not None) and (self._parent is not None)

    @property
    def parent(self) -> '_ConfigObject':
        assert self.has_parent
        return self._parent

    @property
    def key(self) -> str:
        assert self.has_parent
        return self._key

    @property
    def keys(self) -> Tuple[str]:
        return tuple(n.key for n in self.walk_from_root(visit_root=False))

    @property
    def root(self) -> '_ConfigObject':
        node = self
        for node in self.walk_to_root(visit_root=True):
            pass
        return node

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Walk                                                                  #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def walk_to_root(self, visit_root=False):
        current = self
        while current.has_parent:
            yield current
            current = current.parent
        if visit_root:
            yield current

    def walk_from_root(self, visit_root=False):
        yield from reversed(list(self.walk_to_root(
            visit_root=visit_root
        )))

    def walk_descendants(self):
        def _iter(node):
            yield node
            for k in node:
                yield from _iter(node[k])
        return _iter(self)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Children                                                              #
    # - for simplicity we do not allow attribute access to children.        #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def add_child(self, key: str, child: '_ConfigObject') -> '_ConfigObject':
        if not isinstance(child, _ConfigObject):
            raise TypeError(f'child must be an instance of {_ConfigObject.__name__}')
        if child.has_parent:
            raise ValueError(f'child already has a parent, and cannot be added.')
        if key in self._children:
            raise KeyError(f'parent already has child with key: {key}')
        # set details
        child._parent = self
        child._key = key
        self._children[key] = child
        # return the added value
        return child

    def del_child(self, key: str):
        child = self.get_child(key)
        # remove details
        child._parent = None
        child._key = None
        del self._children[key]

    def get_child(self, key: str):
        return self._children[key]

    def has_child(self, key: str):
        return key in self._children

    __setitem__ = add_child
    __getitem__ = get_child
    __contains__ = has_child

    def __iter__(self):
        yield from self._children


# ========================================================================= #
# Group                                                                     #
# ========================================================================= #


class Group(_ConfigObject):

    def __init__(
            self,
            named_nodes: Dict[str, Union['Group', 'Option']] = None,
    ):
        super().__init__()
        # add groups or options
        if named_nodes:
            for k, v in named_nodes.items():
                self.add_child(k, v)
            # we dont need to validate groups because we dont
            # directly support conversion to them yet.

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Override                                                              #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def add_child(self, key, node: Union['Group', 'Option']) -> Union['Group', 'Option']:
        if not isinstance(node, (Group, Option)):
            raise TypeError(f'Can only add {Group.__name__} and {Option.__name__} nodes to {Group.__name__}')
        # add!
        assert super().add_child(key, node) is node
        # we dont return the above because of type warnings
        return node

    # override to yield in order
    def __iter__(self):
        yield from self.groups
        yield from self.options

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Getters                                                               #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @property
    def root(self) -> 'Group':
        g = super().root
        assert isinstance(g, Group)
        return g

    @property
    def groups(self) -> Dict[str, 'Group']:
        return {k: v for k, v in self._children.items() if isinstance(v, Group)}

    @property
    def options(self) -> Dict[str, 'Option']:
        return {k: v for k, v in self._children.items() if isinstance(v, Option)}

    @property
    def group_path(self):
        return '/'.join(self.keys if self.keys else ())

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Groups & Options                                                      #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def get_subgroup(self, key: str) -> 'Group':
        node = self.get_child(key)
        if not isinstance(node, Group):
            raise TypeError(f'node with key: {repr(key)} is not a {Group.__name__}')
        return node

    def del_subgroup(self, key: str) -> 'Option':
        assert self.has_subgroup(key)
        self.del_child(key)

    def get_option(self, key: str) -> 'Option':
        node = self.get_child(key)
        if not isinstance(node, Option):
            raise TypeError(f'node with key: {repr(key)} is not a {Option.__name__}')
        return node

    def add_subgroup(self, key: str, value: 'Group') -> 'Group':
        if not isinstance(value, Group):
            raise TypeError(f'adding node with key: {repr(key)} is not a {Group.__name__}')
        return self.add_child(key, value)

    def add_option(self, key: str, value: 'Option') -> 'Option':
        if not isinstance(value, Option):
            raise TypeError(f'adding node with key: {repr(key)} is not a {Option.__name__}')
        return self.add_child(key, value)

    def del_option(self, key: str) -> 'Option':
        assert self.has_suboption(key)
        self.del_child(key)

    def new_subgroup(self, key: str) -> 'Group':
        return self.add_subgroup(key, Group())

    def new_option(self, key: str, data=None, pkg=None, defaults=None) -> 'Option':
        return self.add_option(key, Option(
            data=data, pkg=pkg, defaults=defaults
        ))

    def has_suboption(self, key):
        if not self.has_child(key):
            return False
        if not isinstance(self.get_child(key), Option):
            return False
        return True

    def has_subgroup(self, key):
        if not self.has_child(key):
            return False
        if not isinstance(self.get_child(key), Group):
            return False
        return True

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Walk                                                                  #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def get_subgroup_recursive(self, keys: List[str], make_missing=False) -> 'Group':
        def _recurse(group: Group, old_keys):
            if not old_keys:
                return group
            key, keys = old_keys[0], old_keys[1:]
            if make_missing:
                if not group.has_subgroup(key):
                    return _recurse(group.new_subgroup(key), keys)
            return _recurse(group.get_subgroup(key), keys)
        return _recurse(self, keys)

    def get_group_from_path(self, path: str, make_missing=False):
        """
        This function checks if a path is relative or absolute.
        If the path is relative, it traverses from the current group.
        If the path is absolute, it traverses from the root group.

        Supports make_missing like get_subgroups_recursive(...)
        """
        group_keys, is_relative = V.split_config_path(path)
        # get the group corresponding to the path - must handle relative & root paths
        root = (self if is_relative else self.root)
        return root.get_subgroup_recursive(group_keys, make_missing=make_missing)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Strings                                                               #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f'{self.__class__.__name__}({self._children})'

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Debug                                                                 #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def debug_print_tree(self, colors=True, show_options=True, full_option_path=True, full_group_path=True):
        from eunomia.config._config_debug import debug_print_tree
        return debug_print_tree(self, colors=colors, show_options=show_options, full_option_path=full_option_path, full_group_path=full_group_path)


# ========================================================================= #
# Option                                                                    #
# ========================================================================= #


class Option(_ConfigObject):

    def __init__(
            self,
            data: dict = None,
            pkg: str = None,
            defaults: list = None,
    ):
        super().__init__()
        # extract components
        self._data = V.validate_option_data(data)
        self._pkg = V.validate_option_package(pkg)
        self._defaults = V.validate_option_defaults(defaults, allow_config_nodes=True)

    @property
    def pkg(self):
        return self._pkg

    @property
    def data(self):
        return self._data.copy()

    @property
    def defaults(self):
        return self._defaults.copy()

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # getters                                                               #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @property
    def root(self) -> 'Group':
        assert self.has_parent
        g = super().root
        assert isinstance(g, Group)
        return g

    @property
    def group(self) -> Group:
        assert self.has_parent
        g = self.parent
        assert isinstance(g, Group)
        return g

    @property
    def group_keys(self) -> Tuple[str]:
        return self.group.keys

    @property
    def group_path(self):
        return self.group.group_path

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # getters - data                                                        #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    class _ReplaceStrings(RecursiveTransformer):
        def _transform_str(self, value):
            return SubNode(value)
        def __transform_default__(self, value):
            return value
        def _transform_dict_key(self, key):
            # do not allow interpolation on keys
            if isinstance(key, str):
                return key
            elif isinstance(key, ConfigNode):
                raise TypeError(f'keys in configs cannot be config nodes: {key}')
            return super()._transform_dict_key(key)

    def get_unresolved_defaults(self):
        return self._ReplaceStrings().transform(self._defaults)

    def get_unresolved_package(self):
        return self._ReplaceStrings().transform(self._pkg)

    def get_unresolved_data(self):
        return self._ReplaceStrings().transform(self._data)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Children - disabled for the option node                               #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    add_child = None
    get_child = None
    __setitem__ = None
    __getitem__ = None
    __iter__ = ().__iter__

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Strings                                                               #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f'{self.__class__.__name__}(data={repr(self._data)}, pkg={repr(self._pkg)}, defaults={repr(self._defaults)})'


# ========================================================================= #
# End                                                                       #
# ========================================================================= #



from typing import Union as _Union
from eunomia.config._config import _ConfigObject, Group, Option


# ========================================================================= #
# debug tree                                                                #
# ========================================================================= #


def debug_print_tree(root: _Union[Group, Option], colors=True, show_options=True, full_option_path=True, full_group_path=True):
    from attr import dataclass

    @dataclass
    class _WalkObj:
        node: _ConfigObject
        visited: bool
        is_last: bool
        @property
        def is_group(self):
            return isinstance(self.node, Group)
        @property
        def has_groups_after(self):
            return self.node.has_parent and isinstance(self.node.parent, Group) and bool(self.node.parent.groups)
        @property
        def key(self):
            return (self.is_group, self.visited, self.is_last, self.has_groups_after)
        def __str__(self):
            return f":{''.join(map(str, map(int, self.key)))}"

    def _is_last_iter(items):
        yield from ((item, i == len(items ) -1) for i, item in enumerate(items))

    def _walk():
        def _recurse(node, is_last, stack):
            stk = stack + [_WalkObj(node, False, is_last)]
            yield stk
            stk[-1].visited = True
            if isinstance(node, Group):
                if show_options:
                    for o, l in _is_last_iter(node.options.values()):
                        yield from _recurse(o, l, stk)
                for g, l in _is_last_iter(node.groups.values()):
                    yield from _recurse(g, l, stk)
        return _recurse(root, True, [])

    if colors:
        nG, nO, S, G, O, R = '\033[35m', '\033[33m', '\033[90m', '\033[95m', '\033[93m', '\033[0m'
    else:
        nG, nO, S, G, O, R = '', '', '', '', '', ''

    TREE = {
        (1, 1, 1, 0): f'   ',           # group,  visited,   last,  has groups after
        (1, 1, 1, 1): f'   ',           # group,  visited,   last,  no groups after
        (1, 0, 1, 0): f'   ',           # group,  unvisited, last,  has groups after
        (1, 1, 0, 1): f' {S}│{R} ',     # group,  visited,   inner, no groups after
        (1, 0, 0, 1): f' {S}├{G}─{R}',  # group,  unvisited, inner, no groups after
        (1, 0, 1, 1): f' {S}╰{G}─{R}',  # group,  unvisited, last,  no groups after
        (0, 0, 0, 1): f' {S}│{R} ',     # option, unvisited, last,  has groups after
        (0, 0, 1, 1): f' {S}├{O}╌{R}',  # option, unvisited, last,  no groups after
        (0, 0, 0, 0): f' {S}├{O}╌{R}',  # option, unvisited, inner, has groups after
        (0, 0, 1, 0): f' {S}╰{O}╌{R}',  # option, unvisited, last,  has groups after
    }

    for stack in _walk():
        (*_, item) = stack
        tree = ''.join(TREE.get(o.key, f'ERR{o}') for o in stack[1:])
        if item.is_group:
            keys = [f'/{k}' for k in (item.node.keys if item.node.keys else ('',))]
            name = f"{nG}{keys[-1]}{R}"
            if full_group_path:
                name = f"{S}{''.join(keys[:-1])}{R}" + name
        else:
            name = f"{nO}{item.node.key}{R}"
            if full_option_path:
                name = f"{S}{('/' + '/'.join(item.node.keys[:-1]))}:{R} " + name
        print(tree, name)


# ========================================================================= #
# End                                                                       #
# ========================================================================= #



import os as _os
import keyword as _keyword
from typing import Union as _Union, Tuple as _Tuple

from eunomia.config import keys as _K


# ========================================================================= #
# Keys                                                                      #
# ========================================================================= #


def validate_identifier(key) -> str:
    if not isinstance(key, str):
        raise TypeError(f'identifier is not a string type: {type(key)}')
    if not key:
        raise ValueError(f'identifier cannot be empty: {repr(key)}')
    if not str.isidentifier(key):
        raise ValueError(f'identifier is not a valid python identifier: {repr(key)}')
    if _keyword.iskeyword(key):
        raise ValueError(f'identifier is a python keyword: {repr(key)}')
    if key in _K.RESERVED_KEYS:
        raise ValueError(f'identifier is eunomia reserved key: {repr(key)}')
    if key.startswith('__') and key.endswith('__'):
        raise ValueError(f'identifier is reserved: {repr(key)}')
    return key


def validate_identifier_list(keys) -> list:
    if not isinstance(keys, list):
        raise TypeError(f'identifier list must be a list type: {type(keys)}')
    for k in keys:
        validate_identifier(k)
    return keys


def validate_config_identifier(key) -> str:
    return validate_identifier(key)


# ========================================================================= #
# Values                                                                    #
# ========================================================================= #


def _split_path(path: str, sep: str) -> (str, bool):
    # check and remove prefix
    is_prefixed = False
    if path.startswith(sep):
        is_prefixed, path = True, path[len(sep):]
    # split path
    if path:
        split = str.split(path, sep)
    else:
        split = []
    # check identifiers
    return validate_identifier_list(split), is_prefixed


def split_package_path(path: str) -> (str, bool):
    if path in {_K.PKG_GROUP, _K.PKG_ROOT}:
        raise RuntimeError('special package keys should be handled separately')
    keys, is_relative = _split_path(path, '.')
    return keys, is_relative


def split_config_path(path: str) -> (str, bool):
    keys, is_not_relative = _split_path(path, '/')
    return keys, not is_not_relative


# ========================================================================= #
# Paths                                                                     #
# ========================================================================= #


def validate_package_path(path) -> str:
    if not isinstance(path, str):
        raise TypeError(f'package path is not of type string: {type(path)}')
    # check for aliases
    if path in {_K.PKG_GROUP, _K.PKG_ROOT}:
        return path
    # try split
    split_package_path(path)
    return path


def validate_config_path(path) -> str:
    if not isinstance(path, str):
        raise TypeError(f'config path is not of type string: {type(path)}')
    # try split
    split_config_path(path)
    return path


# ========================================================================= #
# config values                                                             #
# ========================================================================= #


def validate_option_data(data) -> dict:
    if data is None:
        return {}

    from eunomia.config.nodes import ConfigNode

    if isinstance(data, ConfigNode):
        raise TypeError(f'option data object can never be a config node: {repr(data)}')
    if not isinstance(data, dict):
        raise TypeError(f'option data object must be of type dict: {repr(data)}')
    return _validate_option_data(data)


def _validate_option_data(value):
    if value is None:
        pass
    elif isinstance(value, dict):
        for k, v in value.items():
            validate_identifier(k)
            _validate_option_data(v)
    elif isinstance(value, list):
        for v in value:
            _validate_option_data(v)
    elif isinstance(value, (int, float, str)):
        pass
    else:
        raise TypeError(f'unsupported data value type: {type(value)} with value: {repr(value)}')
    return value


def validate_option_package(pkg) -> str:
    if pkg is None:
        return _K.DEFAULT_PKG

    from eunomia.config.nodes import ConfigNode

    if isinstance(pkg, ConfigNode):
        raise TypeError(f'option package can never be a config node: {repr(pkg)}')
    try:
        return validate_package_path(pkg)
    except Exception as e:
        raise ValueError(f'option package is invalid: {repr(pkg)}').with_traceback(e.__traceback__)


def validate_option_defaults(defaults, allow_config_nodes=False) -> list:
    if defaults is None:
        return []

    from eunomia.config.nodes import ConfigNode

    if isinstance(defaults, ConfigNode):
        raise TypeError(f'option defaults can never directly be a config node, only its items: {repr(defaults)}')
    try:
        return split_defaults_list_items(
            defaults,
            allow_config_node_return=allow_config_nodes
        )
    except Exception as e:
        raise ValueError(f'option defaults is invalid: {repr(defaults)}').with_traceback(e.__traceback__)


def validate_option_type(typ) -> str:
    if typ is None:
        return _K.TYPE_OPTION

    from eunomia.config.nodes import ConfigNode

    if isinstance(typ, ConfigNode):
        raise TypeError(f'option type can never be a config node: {repr(typ)}')
    if not isinstance(typ, str):
        raise TypeError(f'option type must be of type string: {repr(typ)}')
    if typ != _K.TYPE_OPTION:
        raise ValueError(f'option  type must be: {repr(typ)}')
    return typ


# ========================================================================= #
# Options                                                                   #
# ========================================================================= #


def split_defaults_item(item, allow_config_node_return=False) -> _Union[_Tuple[str, str], str, 'Option']:
    from eunomia.config.nodes import ConfigNode
    from eunomia.config._config import Option

    if isinstance(item, ConfigNode):
        raise RuntimeError('single defaults items that are config nodes should be resolved before being split.')

    # support references to options
    # - these need to be resolved outside of this function in case things are
    #   updated while constructing the config tree
    if isinstance(item, Option):
        if allow_config_node_return:
            return item
        raise RuntimeError(f'single defaults items that are an {Option.__name__} instance are disabled.')

    # check the normal types
    if item == _K.OPT_SELF:
        # this must be handled externally
        # its possibly an error if this is used here!
        return item
    elif isinstance(item, (tuple, list)):
        group_path, option_name = item
    elif isinstance(item, str):
        option_name = _os.path.basename(item)
        group_path = _os.path.dirname(item)
    elif isinstance(item, dict):
        assert len(item) == 1
        group_path, option_name = list(item.items())[0]
    else:
        raise TypeError(f'invalid defaults item type: {type(item)} with value: {item}')

    # we allow this to return config nodes,
    # but they need to be resolved
    if not allow_config_node_return:
        if isinstance(group_path, ConfigNode) or isinstance(option_name, ConfigNode):
            raise TypeError('returning of config nodes is not allowed')

    # validate identifiers
    if not isinstance(group_path, ConfigNode):
        group_path = validate_config_path(group_path)
    if not isinstance(option_name, ConfigNode):
        option_name = validate_config_identifier(option_name)

    # return!
    return group_path, option_name


def split_defaults_list_items(defaults, allow_config_node_return=False) -> list:
    if not isinstance(defaults, list):
        raise TypeError(f'defaults list must be of type: {list}')
    return [
        split_defaults_item(item, allow_config_node_return=allow_config_node_return)
        for item in defaults
    ]


def validate_resolved_defaults_item(group_path, option_name):
    from eunomia.config.nodes import ConfigNode

    if isinstance(group_path, ConfigNode) or isinstance(option_name, ConfigNode):
        raise ValueError('group_path and option_name have not been resolved')

    return validate_config_path(group_path), validate_config_identifier(option_name)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



import inspect
import re

from eunomia._util_dict import recursive_getitem, dict_recursive_update
from eunomia.config import keys as K
from eunomia.config import validate as V
from eunomia.config import Option


# ========================================================================= #
# Helper                                                                    #
# ========================================================================= #


def _fn_get_kwargs(func) -> dict:
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }


def _fn_get_args(func) -> list:
    signature = inspect.signature(func)
    return [
        k
        for k, v in signature.parameters.items()
        if v.default is inspect.Parameter.empty
    ]


def _fn_get_all_args(func) -> list:
    signature = inspect.signature(func)
    return list(signature.parameters.keys())


def _fn_get_module_path(obj):
    # get package search paths
    import os
    import sys
    # get module path
    path = os.path.abspath(inspect.getmodule(obj).__file__)
    # return the shortest relative path from all the packages
    rel_paths = []
    for site in sys.path:
        site = os.path.abspath(site)
        if os.path.commonprefix([site, path]) == site:
            rel_paths.append(os.path.relpath(path, site))
    # get shortest rel path
    rel_paths = sorted(rel_paths, key=str.__len__)
    assert len(rel_paths) > 0, f'no valid path found to: {obj}'
    # get the correct path
    path = rel_paths[0]
    assert path.endswith('.py')
    path = path[:-len('.py')]
    return os.path.normpath(path)


def _fn_get_import_path(obj):
    module_path = '.'.join(_fn_get_module_path(obj).split('/'))
    return f'{module_path}.{obj.__name__}'


def _camel_to_snake(name):
    name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()


# ========================================================================= #
# Advanced Group                                                            #
# ========================================================================= #


def make_target_option(
        fn,
        # target function
        target: str = None,
        params: dict = None,
        mode: str = 'any',
        keep_defaults: bool = True,
        # option params extras
        nest_path: str = None,
        data: dict = None,
        pkg: str = None,
        defaults: list = None,
) -> 'Option':
    # get various defaults
    data = {} if data is None else data

    # check nest path
    if nest_path is not None:
        nest_path, is_relative = V.split_package_path(nest_path)
        assert not is_relative, 'nest path must not be relative'
    else:
        nest_path = []

    # get the dictionary to merge the target into
    targ_merge_dict = recursive_getitem(data, nest_path, make_missing=True)
    if not isinstance(targ_merge_dict, dict):
        raise ValueError('nested object in data must be a dictionary that the target can be merged into.')
    if targ_merge_dict:
        raise ValueError('nested object in data must be empty, otherwise target conflicts can occur.')

    # make data for the option
    dict_recursive_update(
        left=targ_merge_dict,
        right=make_target_dict(fn, target=target, params=params, mode=mode, keep_defaults=keep_defaults),
    )

    # make option
    return Option(
        data=data,
        pkg=pkg,
        defaults=defaults
    )


def make_target_dict(
        func,
        # target function
        target: str = None,
        params: dict = None,
        mode: str = 'any',
        keep_defaults: bool = True,
) -> dict:
    # get default values
    target = _fn_get_import_path(func) if target is None else target
    overrides = {} if params is None else params

    # if we should include all the non-overridden default
    # parameters in the final config
    if keep_defaults:
        defaults = _fn_get_kwargs(func)
    else:
        defaults = {}

    # get the parameters that can be overridden
    args, kwargs, all_args = _fn_get_args(func), _fn_get_kwargs(func), _fn_get_all_args(func)
    if mode == 'kwargs':
        allowed_overrides = set(kwargs)
    elif mode == 'any':
        allowed_overrides = set(args) | set(kwargs)
    elif mode == 'all':
        allowed_overrides = set(args) | set(kwargs)
    elif mode == 'unchecked':
        allowed_overrides = None
    else:
        raise KeyError(f'Invalid override mode: {repr(mode)}')

    # generate final list of parameter overrides
    if mode == 'unchecked':
        defaults.update(overrides)
        overrides = {}
    else:
        for k in all_args:  # sorted
            if k in allowed_overrides:
                if k in overrides:
                    defaults[k] = overrides.pop(k)

    # check that no extra unused overrides exist
    # and that _target_ is not a parameter name
    assert not overrides, f'cannot override params: {list(overrides.keys())}'
    assert K.MARKER_KEY_TARGET not in defaults, f'object {target} has conflicting optional parameter: {K.MARKER_KEY_TARGET}'

    # check that nothing is left over according to the override mode
    if mode == 'all':
        missed_args = set(args) - set(defaults)
        if missed_args:
            missed_args = [a for a in all_args if a in missed_args]  # sorted
            raise AssertionError(f'all non-default parameters require an override: {missed_args}')

    # return final dictionary
    return {
        K.MARKER_KEY_TARGET: target,
        **defaults,
    }


# ========================================================================= #
# END                                                                       #
# ========================================================================= #




from typing import Tuple

from eunomia._util_dict import recursive_getitem, dict_recursive_update
from eunomia.config import Option, Group
from eunomia.config.nodes import ConfigNode

from eunomia.config import keys as K
from eunomia.config import validate as V


# ========================================================================= #
# Config Loader                                                             #
# ========================================================================= #


class ConfigLoader(object):

    def __init__(self, root_group: Group):
        if not isinstance(root_group, Group):
            raise TypeError(f'root_group must be a {Group.__name__}')
        self._root_group = root_group
        # merged items
        self._merged_options = None
        self._merged_config = None

    def load_config(self, config_name, return_merged_options=False):
        """
        flatten and merge the options lists using DFS, while
        simultaneously merging the config

        - When config dictionaries are encountered, those are also
          processed using DFS to obtain substituted values, before
          being merged into the config.
            * Keys are not allowed to be substituted values
        """
        self._merged_options = {}
        self._merged_config = {}

        # ===================== #
        # 1. entry point for dfs, get initial option
        root_group = self._root_group
        entry_option = root_group.get_option(config_name)
        # ===================== #
        # 2. perform dfs & merging and finally resolve values
        self._visit_option(entry_option)
        # ===================== #
        # 3. finally resolve all the values in the config
        self._resolve_all_values()
        # ===================== #

        # done, return the result
        if return_merged_options:
            return self._merged_config, self._merged_options
        return self._merged_config

    def _visit_option(self, option: Option):
        # ===================== #
        # 1. check where to process self, and make sure self is in the options list
        # by default we want to merge this before children so we can reference its values.
        defaults = option.get_unresolved_defaults()
        if K.OPT_SELF not in defaults:
            # allow referencing parent values in children
            defaults = [K.OPT_SELF] + defaults
            # # allow parent overwriting children values
            # group_paths = group_paths + [s.OPT_SELF]
        # ===================== #
        # 2. process options in order
        for default_item in defaults:
            # handle different cases
            if default_item == K.OPT_SELF:
                # ===================== #
                # 2.a if self is encountered, merge into config. We skip the
                #     value of the option_name here as it is not needed.
                self._merge_option(option)
                # ===================== #
            else:
                # normalise the thing, can be strings, or dicts
                group_path, option_name = self._resolve_default_item(default_item)
                # ===================== #
                # 2.b dfs through options
                # supports relative & absolute paths
                group = option.group.get_group_from_path(group_path, make_missing=False)
                # visit the next option
                self._visit_option(group.get_option(option_name))
                # ===================== #

    def _resolve_default_item(self, default_item):
        # split the default item
        result = V.split_defaults_item(
            self._resolve_value(default_item),
            allow_config_node_return=True,
        )
        # handle the case where an option instance was in the list
        if isinstance(result, Option):
            group_path, option_name = result.group_path, result.key
        elif result == K.OPT_SELF:
            raise RuntimeError('This is a bug')
        else:
            group_path, option_name = result
        # check that the components are valid
        return V.validate_resolved_defaults_item(
            self._resolve_value(group_path),
            self._resolve_value(option_name),
        )

    def _merge_option(self, option: Option):
        # 1. check that the group has not already been merged
        # 2. check that the option has not already been merged
        self._merge_option_mark_visited(option)
        # 3. merged the data
        self._merge_option_into_config(option)

    def _merge_option_mark_visited(self, option: Option):
        # TODO: this should have different modes
        #       - do not allow from the same group to be merged
        #       - allow from the same group to be merged
        #       1. check that the group has not already been merged
        #       2. check that the option has not already been merged
        # get the path to the config - recursive version of whats listed in the __defaults__
        # maybe lift the non-recursive limitation in future?
        group_keys = option.group_keys
        # check that this is not a duplicate
        if group_keys in self._merged_options:
            prev_added_keys = self._merged_options[group_keys]
            raise KeyError(f'Group has duplicate entry: {repr(group_keys)}. '
                           f'Key previously added by: {repr(prev_added_keys)}. '
                           f'Current config file is: {repr(option.keys)}.')
        # merge the path!
        self._merged_options[group_keys] = option.keys

    def _merge_option_into_config(self, option: Option):
        # 1. get the package and handle special values
        keys = self._resolve_package(option)
        # 2. get the root config object according to the package
        root = recursive_getitem(self._merged_config, keys, make_missing=True)
        # 3. merge the option into the config
        dict_recursive_update(left=root, right=option.get_unresolved_data(), safe_merge=True)

    def _resolve_value(self, value):
        # 1. allow interpolation of config objects
        # 2. TODO: process dictionary syntax with _node_ keys
        if isinstance(value, ConfigNode):
            value = value.get_config_value(self._merged_config, self._merged_options, {})
        return value

    def _resolve_package(self, option) -> Tuple[str]:
        path = self._resolve_value(option.get_unresolved_package())
        # check the type
        if not isinstance(path, str):
            raise TypeError(f'{K.KEY_PKG} must be a string')
        # handle special values
        if path == K.PKG_ROOT:
            keys = ()
        elif path == K.PKG_GROUP:
            keys = option.group_keys
        else:
            keys, is_relative = V.split_package_path(path)
            if is_relative:
                keys = option.group_keys + keys
        # return the keys
        return keys

    def _resolve_all_values(self):
        # TODO: this is wrong!
        # TODO: config is not mutated on the fly, cannot substitute chains of values.
        self._merged_config = ConfigNode.recursive_get_config_value(
            self._merged_config,
            self._merged_options,
            {},
            self._merged_config
        )


# ========================================================================= #
# End                                                                       #
# ========================================================================= #



import pytest
from eunomia.config import keys as K, Option, Group
from eunomia.config import validate as V

# ========================================================================= #
# Test YAML & Custom Tags                                                   #
# ========================================================================= #


def _do_identifier_tests(validator, prefix=''):
    # check simple valid
    validator(prefix + 'valid_id')
    with pytest.raises(ValueError):  # TODO: fix , match="is_not_special_key"):
        validator(prefix + '__type__')
    with pytest.raises(ValueError):  # TODO: fix , match="is_not_reserved"):
        validator(prefix + '__invalid_id__')
    validator(prefix + '_okay_id_')
    with pytest.raises(ValueError):  # TODO: fix , match="is_identifier"):
        validator(prefix + '1234')
    with pytest.raises(ValueError):  # TODO: fix , match="is_not_keyword"):
        validator(prefix + 'with')
    with pytest.raises(TypeError):  # TODO: fix , match="should be instance of"):
        validator(1234)


def test_identifier():
    _do_identifier_tests(V.validate_identifier)


def test_pkg_path():
    # check - absolute
    assert V.validate_package_path('') == ''
    assert V.validate_package_path('valid1') == 'valid1'
    assert V.validate_package_path('valid1.valid2') == 'valid1.valid2'
    with pytest.raises(ValueError): V.validate_package_path('valid1.')
    # check - relative
    assert V.validate_package_path('.') == '.'
    assert V.validate_package_path('.valid1') == '.valid1'
    assert V.validate_package_path('.valid1.valid2') == '.valid1.valid2'
    with pytest.raises(ValueError): V.validate_package_path('.valid1.')

    # check splits - absolute
    assert V.split_package_path('') == ([], False)
    assert V.split_package_path('valid1') == (['valid1'], False)
    assert V.split_package_path('valid1.valid2') == (['valid1', 'valid2'], False)
    with pytest.raises(ValueError): V.split_package_path('valid1.')
    # check splits - relative
    assert V.split_package_path('.') == ([], True)
    assert V.split_package_path('.valid1') == (['valid1'], True)
    assert V.split_package_path('.valid1.valid2') == (['valid1', 'valid2'], True)
    with pytest.raises(ValueError): V.split_package_path('.valid1.')

    # check identifiers, should be same as above
    _do_identifier_tests(V.validate_package_path)
    _do_identifier_tests(V.validate_package_path, prefix='valid1.')
    _do_identifier_tests(V.validate_package_path, prefix='_marker_.')
    with pytest.raises(ValueError):  # TODO: fix , match='is_not_reserved'):
        _do_identifier_tests(V.validate_package_path, prefix='__invalid1__.')

    # check group...
    with pytest.raises(ValueError):  # TODO: fix , match='is_identifier'):
        V.validate_package_path('valid1/valid2/valid3')

    # check special keys
    V.validate_package_path('<root>')
    V.validate_package_path('<group>')
    # split special keys
    with pytest.raises(RuntimeError): V.split_package_path('<root>')
    with pytest.raises(RuntimeError): V.split_package_path('<group>')


def test_group_path():
    # check - relative
    assert V.validate_config_path('') == ''
    assert V.validate_config_path('valid1') == 'valid1'
    assert V.validate_config_path('valid1/valid2') == 'valid1/valid2'
    with pytest.raises(ValueError): V.validate_config_path('valid1/')
    # check - absolute
    assert V.validate_config_path('/') == '/'
    assert V.validate_config_path('/valid1') == '/valid1'
    assert V.validate_config_path('/valid1/valid2') == '/valid1/valid2'
    with pytest.raises(ValueError): V.validate_config_path('/valid1/')

    # check splits - relative
    assert V.split_config_path('') == ([], True)
    assert V.split_config_path('valid1') == (['valid1'], True)
    assert V.split_config_path('valid1/valid2') == (['valid1', 'valid2'], True)
    with pytest.raises(ValueError): V.split_config_path('valid1/')
    # check splits - absolute
    assert V.split_config_path('/') == ([], False)
    assert V.split_config_path('/valid1') == (['valid1'], False)
    assert V.split_config_path('/valid1/valid2') == (['valid1', 'valid2'], False)
    with pytest.raises(ValueError): V.split_config_path('/valid1/')

    # check identifiers, should be same as above
    _do_identifier_tests(V.validate_config_path)
    _do_identifier_tests(V.validate_config_path, prefix='valid1/')
    _do_identifier_tests(V.validate_config_path, prefix='_marker_/')
    with pytest.raises(ValueError):  # TODO: fix , match='is_not_reserved'):
        _do_identifier_tests(V.validate_config_path, prefix='__invalid1__/')

    # check group...
    with pytest.raises(ValueError):  # TODO: fix , match='is_identifier'):
        V.validate_config_path('valid1.valid2.valid3')

    # check special keys
    with pytest.raises(ValueError): V.validate_config_path('<root>')
    with pytest.raises(ValueError): V.validate_config_path('<group>')
    # split special keys
    with pytest.raises(RuntimeError): V.split_package_path('<root>')
    with pytest.raises(RuntimeError): V.split_package_path('<group>')


def test_defaults():

    # check self
    assert V.split_defaults_item('<self>') == '<self>'
    with pytest.raises(ValueError):
        V.split_defaults_item(['<self>s'])

    # check strings
    assert V.split_defaults_item('option1') == ('', 'option1')
    assert V.split_defaults_item('/option1') == ('/', 'option1')
    assert V.split_defaults_item('group/option1') == ('group', 'option1')
    assert V.split_defaults_item('/group/option1') == ('/group', 'option1')
    with pytest.raises(ValueError):
        V.split_defaults_item('/1invalidgroup/option1')

    # check dictionaries
    assert V.split_defaults_item({'': 'option1'}) == ('', 'option1')
    assert V.split_defaults_item({'/': 'option1'}) == ('/', 'option1')
    assert V.split_defaults_item({'group': 'option1'}) == ('group', 'option1')
    assert V.split_defaults_item({'/group': 'option1'}) == ('/group', 'option1')
    with pytest.raises(ValueError):
        V.split_defaults_item({'/group/subgroup': '1invalidopt'})
    with pytest.raises(ValueError):
        V.split_defaults_item({'/1invalidgroup/subgroup': 'option1'})

    # check tuples
    assert V.split_defaults_item(('', 'option1')) == ('', 'option1')
    assert V.split_defaults_item(('/', 'option1')) == ('/', 'option1')
    assert V.split_defaults_item(('group', 'option1')) == ('group', 'option1')
    assert V.split_defaults_item(('/group', 'option1')) == ('/group', 'option1')

    assert V.split_defaults_item(['', 'option1']) == ('', 'option1')
    assert V.split_defaults_item(['/', 'option1']) == ('/', 'option1')
    assert V.split_defaults_item(['group', 'option1']) == ('group', 'option1')
    assert V.split_defaults_item(['/group', 'option1']) == ('/group', 'option1')

    # check options as refs
    option = Group().new_option('default')
    with pytest.raises(RuntimeError, match='single defaults items that are an Option instance are disabled'):
        V.split_defaults_list_items([option])
    with pytest.raises(RuntimeError, match='single defaults items that are an Option instance are disabled'):
        V.split_defaults_list_items([option], allow_config_node_return=False)
    assert V.split_defaults_list_items([option], allow_config_node_return=True) == [option]

    # check lots
    assert V.split_defaults_list_items([]) == []
    assert V.split_defaults_list_items([
        'group1/option1',
        '/group2/option2',
        {'/group/subgroup': 'option1'},
        {'group/subgroup2': 'option2'},
        '<self>',
        option,
    ], allow_config_node_return=True) == [
        ('group1', 'option1'),
        ('/group2', 'option2'),
        ('/group/subgroup', 'option1'),
        ('group/subgroup2', 'option2'),
        '<self>',
        option,
    ]

